#include <iostream>
#include <syslog.h>
#include "Proxy.h"
#include "Globals.h"
#include "Filter.h"
#include <libpq-fe.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>


using namespace std;

char *patterns[MAXPATTERNS];
int lastpattern = -1;

char *configfile = NULL;
char *forwardfile = NULL;
int debug = 0;
int quiet = 1;
int tfactor = 1;
int potential_add;
int logfacility = LOG_DAEMON;
char hostname[MAXHOST] = "144.122.71.203,::1";
char hostport[MAXPORT] = "8000";
char *default_configfile = "/etc/www6to4.conf";
struct gateway gw_default = {NULL, NULL};
	
		
struct file_list *current_forwardfile;
int (*loaders[NLOADERS]) (client_state *);
	

/** Global Variable For the Packet Data */
char *sourceIP; 

char CFAIL[] = "HTTP/1.0 503 Connect failed\r\n"
"Content-Type: text/html\r\n\r\n"
"<html>\r\n"
"<head>\r\n"
"<title>NetCheck Gateway: Connect failed</title>\r\n"
"</head>\r\n"
"<body>\r\n"
"<h1><center>"
BANNER
"</center></h1>\r\n"
"TCP connection to '%s' failed: %s.\r\n<br>"
"</body>\r\n"
"</html>\r\n";

char CNXDOM[] = "HTTP/1.0 404 Non-existent domain\r\n"
"Content-Type: text/html\r\n\r\n"
"<html>\r\n"
"<head>\r\n"
"<title>NetCheck Gateway: Non-existent domain</title>\r\n"
"</head>\r\n"
"<body>\r\n"
"<h1><center>"
BANNER
"</center></h1>\r\n"
"No such domain: %s\r\n"
"</body>\r\n"
"</html>\r\n";

char BLOCK[] = "<body>\r\n"
"<h1><center>"
BANNER
"</center></h1>\r\n"
"Bad Content: %s\r\n"
"</body>\r\n"
"</html>\r\n";



char CHEADER[] = "HTTP/1.0 400 Invalid header received from browser\r\n\r\n";
char SHEADER[] = "HTTP/1.0 502 Invalid header received from server\r\n\r\n";
char CSUCCEED[] = "HTTP/1.0 200 Connection established\r\n"
"Proxy-Agent: NetCheck/" "\r\n\r\n";

/** prototypes */
	/** proxy */
char *safe_strerror (int err);

	/** parsers */
int strcmpic (char *s1, char *s2);
int strncmpic (char *s1, char *s2, size_t n);
int add_to_iob (struct client_state *csp, char *buf, int n);
char *get_header (struct client_state *csp);
void free_http_request (struct http_request *http);
void parse_http_request (char *req, struct http_request *http, struct client_state *csp);
int ssplit (char *s, char *c, char *v[], int n, int m, int l);
struct gateway *forward_url (struct http_request *http, struct client_state *csp);
struct url_spec dsplit (char *domain);
int domaincmp (struct url_spec * pattern, struct url_spec * fqdn);
int parse_content_length(char *p);
void enlist (struct list * h, char *s);
void destroy_list (struct list * h);
char *list_to_text (struct list * h);

	/** loaders */
int loaders_strncmpic(const char *s1, const char *s2, size_t n);
void unload_forwardfile (struct forward_spec *b);
int load_forwardfile (struct client_state *csp);
char *strsav (char *old, char *text_to_append);
void add_loader (int (*loader) ());
int run_loader (struct client_state *csp);

	/** socket */
int open_server_socket (char *hostname, char *port, int af);
int accept_conn (int *fds);
int direct_connect (struct gateway *gw, struct http_request *http, struct client_state *csp);
int connect_to (char *host, char *port, struct client_state * csp);
	
Proxy::Proxy() {
		
		lastpattern = -1;

	
		configfile = NULL;
		forwardfile = NULL;
		debug = 0;
		quiet = 1;
		tfactor = 1;
		potential_add;
		logfacility = LOG_DAEMON;
	
		

// 		 Globals::packetParser.packet.destinationIP=(char *)malloc(sizeof(char)*NI_MAXHOST);	
// 		 Globals::packetParser.packet.sourceIP=(char *)malloc(sizeof(char)*NI_MAXHOST);

}

bool isSamePacket(NetworkTrafficLog packet,NetworkTrafficLog secondPacket){

	if (((!strcmp(packet.sourceIP,secondPacket.sourceIP)) &&
		      (!strcmp(packet.destinationIP,secondPacket.destinationIP))) ||
		      ((!strcmp(packet.sourceIP,secondPacket.destinationIP)) &&
		      (!strcmp(packet.destinationIP,secondPacket.sourceIP))))
		return true;
	return false;

	/**Eger sourceIP disaridan gelen bir paket icin belli deilse bunu kullanmak zorunda kalabilirim */
	// 		      (packet.sourcePort==networkTrafficLog.at(i).destinationPort) &&
	// 		      (packet.destinationPort==networkTrafficLog.at(i).sourcePort)) ){
		
	
}

/**NESLI*/
/**This function is responsible from the storage of the packets on the Newtwork Traffic Log*/
void DumpIntoNetworkLog(NetworkTrafficLog packet){

	
	struct tm *current;
	time_t ntime=time(NULL);
	int inMinute;
	
		
	PGresult *result; 
	PGconn *connection;
	char query[1000000];
			
	const char *pghost;
	const char *pgport;
	const char *pgoptions;
	const char *pgtty;
	const char *dbname;	
	const char *login;
	const char *pwd;
	
	pghost="localhost";
	pgport="5432";
	pgoptions=NULL;
	pgtty=NULL;
	dbname="filiz";
	login="postgres";
	pwd="";

	connection = PQsetdbLogin(pghost,pgport,pgoptions,pgtty,dbname,login,pwd);

	if (PQstatus(connection) != CONNECTION_OK){
		printf("Baglanti Hatasi. %s",PQerrorMessage(connection));
		PQfinish(connection);
		exit(1);
	}
	
		time(&ntime);
		current=localtime(&ntime);
		inMinute=current->tm_hour*60+current->tm_min+current->tm_sec/60; //To see if half an hour has passed
		
				
				
			sprintf(query,"INSERT INTO NetworkTrafficLog(sourceIP,destinationIP,destinationURL,sourcePort,destinationPort,packetSize,firstDay,firstMonth,firstYear,firstHour,firstMinute,firstSecond,lastDay,lastMonth,lastYear,lastHour,lastMinute,lastSecond) VALUES('%s','%s','%s','%d','%d','%d','%d','%d','%d','%d','%d','%d','%d','%d','%d','%d','%d','%d')",packet.sourceIP,packet.destinationIP,packet.accessedURL,packet.sourcePort,packet.destinationPort,packet.packetSize,packet.firstPacketTime.day,packet.firstPacketTime.month,packet.firstPacketTime.year,packet.firstPacketTime.hour,packet.firstPacketTime.minute,packet.firstPacketTime.second,packet.lastPacketTime.day,packet.lastPacketTime.month,packet.lastPacketTime.year,packet.lastPacketTime.hour,packet.lastPacketTime.minute,packet.lastPacketTime.second);
				
			//cout<<"QUERY: "<<query<<endl;
			result=PQexec(connection,query);
			//cout<<"Database Result: "<<result<<endl;
	
		PQfinish(connection);

}
	
	
	
//hande.....................	
void DumpIntoFilteredContent(NetworkTrafficLog packet){

	
	struct tm *current;
	time_t ntime=time(NULL);
	int inMinute;
	
	PGresult *result; 
	PGconn *connection;
	char query[1000000];
			
	const char *pghost;
	const char *pgport;
	const char *pgoptions;
	const char *pgtty;
	const char *dbname;	
	const char *login;
	const char *pwd;
	
	pghost="localhost";
	pgport="5432";
	pgoptions=NULL;
	pgtty=NULL;
	dbname="filiz";
	login="postgres";
	pwd="";

	connection = PQsetdbLogin(pghost,pgport,pgoptions,pgtty,dbname,login,pwd);

	if (PQstatus(connection) != CONNECTION_OK){
		printf("Baglanti Hatasi. %s",PQerrorMessage(connection));
		PQfinish(connection);
		exit(1);
	}
	
		time(&ntime);
		current=localtime(&ntime);
		inMinute=current->tm_hour*60+current->tm_min+current->tm_sec/60; //To see if half an hour has passed
		
				
				
			sprintf(query,"INSERT INTO filteredcontent(sourceIP,destinationIP,destinationURL,firstDay,firstMonth,firstYear,firstHour,firstMinute,firstSecond) VALUES('%s','%s','%s','%d','%d','%d','%d','%d','%d')",packet.sourceIP,packet.destinationIP,packet.accessedURL,packet.firstPacketTime.day,packet.firstPacketTime.month,packet.firstPacketTime.year,packet.firstPacketTime.hour,packet.firstPacketTime.minute,packet.firstPacketTime.second);
				
			result=PQexec(connection,query);
			//cout << "filtered content:" << query;
/*--------------------------------- create table sql -----------------------------------------------


CREATE TABLE filteredcontent (
    filterid serial NOT NULL,
    sourceip character(15) NOT NULL,
    destinationip character(15) NOT NULL,
    destinationurl character(150),
    firstday integer,
    firstmonth integer,
    firstyear integer,
    firsthour integer,
    firstminute integer,
    firstsecond integer,
    primary key(filterid)
);
--------------------------------------------------------------------------------------------------------*/
	
		PQfinish(connection);

}
		
	
	

void Proxy::Tokenize(const string& str, vector<string>& tokens, const string& delimiters)
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

void Proxy::parseAcceptEncoding(char **header,bool *acceptEncodingFlag) {
	
	if(strstr(*header,"POST")){
		*acceptEncodingFlag=false;
		return;
	}

	
	string tempHeader=*header;
	vector <string> tokens;
	Tokenize(tempHeader, tokens, "\r\n");

	vector<string>::iterator it;
  	for( it = tokens.begin(); it != tokens.end(); ++it)
  	{
    		if( it->find("Accept-Encoding:", 0) == 0 )
		{
		    //NESLIHANcout<<endl<<"Accept-Encoding: "<<*it<<endl;
		    *it="Accept-Encoding: ";
		    *acceptEncodingFlag=true; 	
    		}
    
  	}

	tempHeader="";
  	for( it = tokens.begin(); it != tokens.end(); ++it)
  	{
		tempHeader+=*it+"\r\n";
  	}
	tempHeader += "\r\n";

	//NESLIHANcout<<"inside accept encoding parser: "<<tempHeader<<endl;

	*header=(char *)tempHeader.c_str();
}


void parseBody(char *buf,string &sendNetCheck){

	string head = "</head>";
	int loc;

	sendNetCheck=buf;
	
	if((loc = sendNetCheck.find(head, 0))!= string::npos){
		loc += 8;
		sendNetCheck=sendNetCheck.substr(0,loc);
	}
	else
          sendNetCheck="";

	//cout<<"SendNetCheck:(code) "<<sendNetCheck<<endl;
	//cout<<"Buf: "<<buf<<endl;

}

void Proxy::chat (struct client_state * csp,char *sourceIP) {
    char buf[BUFSIZ], *hdr, *p, *req;
    fd_set rfds;
    int i,copy, n, maxfd, server_body;
    struct gateway *gw;
    struct http_request *http;
    struct timeval timeout;
    int bodylen, bodyrecv;
    int sent, length;
    char *bodybuf;
    bool acceptEncodingFlag;
    char *destinationIP,*accessedURL;
    struct hostent *hp;
    struct in_addr *hostaddr;
    struct tm *current;
    time_t ntime;
    vector<NetworkTrafficLog> packets;
    NetworkTrafficLog packet;
    string sendNetCheck,stringBlock=BLOCK; 
    double lastInMinute,firstInMinute;
    double timePassed=0.3;
    bool firstTime=true,firstTime2;		
    
    current=(struct tm *)malloc(sizeof(struct tm));
    ntime = time(NULL);
	

    
     bool doIFilter=false;	
     Filter filter;
    http = csp->http;
    
    

    for (;;) {
        FD_ZERO (&rfds);
        FD_SET (csp->cfd, &rfds);
	
	/**!!!!!!!!!!!!!!!!!!NESLIHAN BU TIMEOUT SANKIM DEGISMELI !!!!!!!!!!!!!!!!!*/
        timeout.tv_sec = 10;
        timeout.tv_usec = 0;

        n = select (csp->cfd + 1, &rfds, NULL, NULL, &timeout);
        if (!n) {
            syslog (LOG_NOTICE, "Timeout reading from client");
	    printf("Timeout reading from client");
            return;
        }
        if (n < 0) {
            syslog (LOG_ERR, "%s", safe_strerror (errno));
            return;
        }
	/**Reads From the Client */
	n = read (csp->cfd, buf, sizeof (buf));
		
	if (n <= 0)
            break;
	/**Time Lookup Should be Computed at Each read/write on cfd */
	time(&ntime);
	current=localtime(&ntime);
	
// 	cout<<"Read From cfd(0): "<<buf<<endl<<"finito"<<endl;	
	/** Accept-Encoding */
	char *pbuf=buf;
	acceptEncodingFlag=false;
	parseAcceptEncoding(&pbuf,&acceptEncodingFlag);
	//NESLIHANcout<<endl<<endl<<"this is my pbuf: "<<pbuf<<endl;
	
	if(acceptEncodingFlag) {
		for(copy=0;copy<BUFSIZ && *pbuf ; copy++)
			buf[copy]=pbuf[copy];
		buf[copy]='\0';	
	}
	
	//NESLIHANcout<<"Ilk buffer: "<<buf<<endl<<" this is my pid: "<<getpid()<<endl;
        add_to_iob (csp, buf, n);

        req = get_header (csp);
        if (!req)
            break;                /* no HTTP request! */
        if (*req == '\0')
            continue;                /* more to come! */

        parse_http_request (req, http, csp);
        
	/**Read the request from the client and Change the packet */
	accessedURL=strdup(http->host);
	hp=gethostbyname(accessedURL);
	hostaddr =(struct in_addr *) hp->h_addr;
	destinationIP=strdup(inet_ntoa(*hostaddr));
	//cout<<"Buraya dikkat edelim lutfen: "<<inet_ntoa(*hostaddr)<<endl;

	/**BEFORE FORK MONITORING*/
// 	pthread_mutex_lock(&(Globals::mutexPacket));
// 	
// 	Globals::packetParser.packet.accessedURL=strdup(http->host);
// 	Globals::packetParser.packet.sourceIP=strdup(sourceIP);
// 	Globals::packetParser.packet.destinationIP=strdup(destinationIP);
// 	Globals::packetParser.packet.packetSize=n;
// 	
// 	
// 	pthread_mutex_unlock(&(Globals::mutexPacket));
// 	
// 	cout<<"After New Packet: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
// 	
// 	cout<<"A packet(0): "<<endl<<"Accessed URI: "<<Globals::packetParser.packet.accessedURL<<endl;
// 	cout<<"Source IP: "<<Globals::packetParser.packet.sourceIP<<endl;
// 	cout<<"Destination IP: "<<Globals::packetParser.packet.destinationIP<<endl;
// 	cout<<"packetSize: "<<Globals::packetParser.packet.packetSize<<endl;
// 	
// 	cout<<"Entering Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
	
	
	if(firstTime){	
	
	//cout<<"Ilk kez: "<<endl<<"Accessed URI: "<<packet.accessedURL<<endl;
	//cout<<"Source IP: "<<packet.sourceIP<<endl;
	//cout<<"Destination IP: "<<packet.destinationIP<<endl;
	//cout<<"packetSize: "<<packet.packetSize<<endl;
		
		
		
	packet.accessedURL=strdup(http->host);
	packet.sourceIP=strdup(sourceIP);
	packet.destinationIP=strdup(destinationIP);
	packet.packetSize=n;

	packet.firstPacketTime.day=current->tm_mday;
	packet.firstPacketTime.month=current->tm_mon+1;
	packet.firstPacketTime.year=current->tm_year+1900;
	packet.firstPacketTime.hour=current->tm_hour;
	packet.firstPacketTime.minute=current->tm_min;
	packet.firstPacketTime.second=current->tm_sec;
				
	packet.lastPacketTime.day=current->tm_mday;
	packet.lastPacketTime.month=current->tm_mon+1;
	packet.lastPacketTime.year=current->tm_year+1900;
	packet.lastPacketTime.hour=current->tm_hour;
	packet.lastPacketTime.minute=current->tm_min;
	packet.lastPacketTime.second=current->tm_sec;
	firstTime=false;
	
	DumpIntoNetworkLog(packet);
	}
	//cout<<"firstLLLKKK: "<<packet.firstPacketTime.hour<<":"<<packet.firstPacketTime.minute<<":"<<packet.firstPacketTime.second<<endl;
					
	//cout<<"Last PAcket Time (1)"<<endl;
	
	
	//cout<<"A packet(0): "<<endl<<"Accessed URI: "<<packet.accessedURL<<endl;
	//cout<<"Source IP: "<<packet.sourceIP<<endl;
	//cout<<"Destination IP: "<<packet.destinationIP<<endl;
	//cout<<"packetSize: "<<packet.packetSize<<endl;
	
	
// 	DumpIntoNetworkLog(Globals::packetParser.packet);
// 	cout<<"Dumpdan Donerkene: "<<Globals::packetParser.networkTrafficLog.size()<<"pid: "<<getpid()<<endl;
	freez (req);
        break;
    }
	
    if (!http->cmd) {
        strcpy (buf, CHEADER);
        printf("Find me: %s--Ent\n",buf);
	write (csp->cfd, buf, strlen (buf));
        return;
    }

    /* decide how to route the HTTP request */
    syslog (LOG_INFO, "Requesting: %s\n", http->cmd);


    
//     for(int hostCopy=0;http->host[hostCopy] && hostCopy<256 ; hostCopy++){	
// 	     Globals::packetParser.packet.accessedURL[hostCopy]=http->host[hostCopy];
// 	     Globals::packetParser.packet.accessedURL[hostCopy+1] = '\0';
//     }
    gw = forward_url (http, csp);

    potential_add = 0;
    for (i = lastpattern; i >= 0; i--) {
        if (strstr(http->path, patterns[i])) {
            if (debug)
            {
		syslog (LOG_INFO, "Pattern: *%s*", patterns[i]);
	        //printf("Requesting: %s", http->cmd);
	    }
            potential_add = 1;
            break;
        }
    }

    /* if forwarding, just pass the request as is */
    if (gw->forward_host)
        enlist (csp->headers, http->cmd);
    else if (!http->ssl) {

            p = NULL;
            p = strsav (p, http->gpc);
            p = strsav (p, " ");
            p = strsav (p, http->path);
            p = strsav (p, " ");
            p = strsav (p, http->ver);
            enlist (csp->headers, p);
            freez (p);
    }

    /* grab the rest of the client's headers */
    bodylen = 0;
    for (;;) {
        p = get_header (csp);
        if (p && (*p == '\0')) {
            n = read (csp->cfd, buf, sizeof (buf));
            if (n < 0) {
                syslog (LOG_ERR, "%s", safe_strerror (errno));
		printf("%s", safe_strerror (errno));
                return;
            }
	    
	    time(&ntime);
	    current=localtime(&ntime);
		
// 	    cout<<"Read From cfd(1): "<<buf<<endl<<"finito"<<endl;
	    /**BEFORE FORK */
// 	pthread_mutex_lock(&(Globals::mutexPacket));
// 	    
//         Globals::packetParser.packet.sourceIP=strdup(sourceIP);
// 	Globals::packetParser.packet.accessedURL=strdup(accessedURL);
// 	Globals::packetParser.packet.destinationIP=strdup(destinationIP);
// 	Globals::packetParser.packet.packetSize=n;
// 	
// 	pthread_mutex_unlock(&(Globals::mutexPacket));
// 	cout<<"A packet(1): "<<endl<<"Accessed URI: "<<Globals::packetParser.packet.accessedURL<<endl;
// 	cout<<"Source IP: "<<Globals::packetParser.packet.sourceIP<<endl;
// 	cout<<"Destination IP: "<<Globals::packetParser.packet.destinationIP<<endl;
// 	cout<<"packetSize: "<<Globals::packetParser.packet.packetSize<<endl;
// 	
	    
// 	cout<<"Entering Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
// 	DumpIntoNetworkLog(current);
// 	cout<<"Exiting Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
	    
	    packet.packetSize+=n;
	    
	    packet.lastPacketTime.day=current->tm_mday;
	    packet.lastPacketTime.month=current->tm_mon+1;
	    packet.lastPacketTime.year=current->tm_year+1900;
	    packet.lastPacketTime.hour=current->tm_hour;
	    packet.lastPacketTime.minute=current->tm_min;
	    packet.lastPacketTime.second=current->tm_sec;
	
	  //  cout<<"Last: "<<packet.lastPacketTime.hour<<":"<<packet.lastPacketTime.minute<<":"<<packet.lastPacketTime.second<<endl;
	    //cout<<"first: "<<packet.firstPacketTime.hour<<":"<<packet.firstPacketTime.minute<<":"<<packet.firstPacketTime.second<<endl;
		
	    firstInMinute=packet.firstPacketTime.hour*60+packet.firstPacketTime.minute+packet.firstPacketTime.second/60;
	    lastInMinute=packet.lastPacketTime.hour*60+packet.lastPacketTime.minute+packet.lastPacketTime.second/60;			
	  //  cout<<"Bura2: "<<lastInMinute-firstInMinute<<endl;
	    if(lastInMinute-firstInMinute>=timePassed)	/**Write to the Database after a minute passes */
	    {
		    DumpIntoNetworkLog(packet);
		    packet.packetSize=0;	
		    timePassed+=5;
	    }
	
	    //cout<<"Packets New Size:(1) "<<packet.packetSize<<endl;
	    
	char *pbuf=buf;
	acceptEncodingFlag=false;
	parseAcceptEncoding(&pbuf,&acceptEncodingFlag);
	//cout<<endl<<endl<<"this is my pbuf: "<<pbuf<<endl;
	if(acceptEncodingFlag) {
		for(copy=0;copy<BUFSIZ && *pbuf ; copy++)
			buf[copy]=pbuf[copy];
		buf[copy]='\0';	
	}
        add_to_iob (csp, buf, n);
        continue;
  }//if

        /* get_header returned NULL, means empty line */
        if (!p)
            break;

        enlist (csp->headers, p);

        /* Check for Content-length. If a content-length header is
           present we have content data that needs to be read as well */
        if (!strncasecmp ("Content-length", p, 14)) {        /* found */
            bodylen = parse_content_length(p);
            syslog (LOG_INFO, "Content-length: %d", bodylen);
	    //printf("Content-length: %d", bodylen);
        }

        /* Finally, free up p */
        freez (p);
    }

    /* add the blank line at the end of the header */
    enlist (csp->headers, "");

    /* make hdr out of list */
    hdr = list_to_text (csp->headers);
    destroy_list (csp->headers);
    length = strlen(hdr);

    /* Do we need to read the body as well? */
    if (bodylen > 0) {                /* Yepp! */
        bodybuf = (char *)calloc(bodylen + length, 1);

        memcpy(bodybuf, hdr, length);

        /* Set hdr to be the new buffer */
        freez (hdr);
        hdr = bodybuf;

        bodybuf = hdr+length;

        bodyrecv = csp->iob->eod - csp->iob->cur;
        memcpy(bodybuf, csp->iob->cur, bodyrecv);

        while (bodyrecv < bodylen) {
            n = read (csp->cfd, bodybuf+bodyrecv, bodylen-bodyrecv);
            
	    if (n < 0) {
                syslog (LOG_ERR, "%s", safe_strerror (errno));
		printf("%s", safe_strerror (errno));
                return;
            }
	    /**Time Lookup Should be Computed at Each read/write on cfd */
	    time(&ntime);
	    current=localtime(&ntime);
	
// 	    pthread_mutex_lock(&(Globals::mutexPacket));
// 	    cout<<"Buffer Read From cfd(2): "<<bodybuf+bodyrecv<<endl<<"finito"<<endl;
        /*    Globals::packetParser.packet.sourceIP=strdup(sourceIP);
	    Globals::packetParser.packet.accessedURL=strdup(accessedURL);
	    Globals::packetParser.packet.destinationIP=strdup(destinationIP);
	    Globals::packetParser.packet.packetSize=n;
		
	    pthread_mutex_unlock(&(Globals::mutexPacket));
	    
	    cout<<"A packet(2): "<<endl<<"Accessed URI: "<<Globals::packetParser.packet.accessedURL<<endl;
	    cout<<"Source IP: "<<Globals::packetParser.packet.sourceIP<<endl;
	    cout<<"Destination IP: "<<Globals::packetParser.packet.destinationIP<<endl;
	    cout<<"packetSize: "<<Globals::packetParser.packet.packetSize<<endl;
	*/
	    
	    packet.packetSize+=n;
	
	
	    packet.lastPacketTime.day=current->tm_mday;
	    packet.lastPacketTime.month=current->tm_mon+1;
	    packet.lastPacketTime.year=current->tm_year+1900;
	    packet.lastPacketTime.hour=current->tm_hour;
	    packet.lastPacketTime.minute=current->tm_min;
	    packet.lastPacketTime.second=current->tm_sec;
	
	  //  cout<<"Last: "<<packet.lastPacketTime.hour<<":"<<packet.lastPacketTime.minute<<":"<<packet.lastPacketTime.second<<endl;
	    //cout<<"first: "<<packet.firstPacketTime.hour<<":"<<packet.firstPacketTime.minute<<":"<<packet.firstPacketTime.second<<endl;
		
	    //cout<<"Last PAcket Time (3)"<<endl;
	    //cout<<"Packets New Size:(2) "<<packet.packetSize<<endl;
	    
	    
	    firstInMinute=packet.firstPacketTime.hour*60+packet.firstPacketTime.minute+packet.firstPacketTime.second/60;
	    lastInMinute=packet.lastPacketTime.hour*60+packet.lastPacketTime.minute+packet.lastPacketTime.second/60;		
	    
	   // cout<<"Bura3: "<<lastInMinute-firstInMinute<<endl;
	    
	    if(lastInMinute-firstInMinute>=timePassed)	/**Write to the Database after a minute passes */
	    {
		    DumpIntoNetworkLog(packet);
		    packet.packetSize=0;
		    timePassed+=5;
	    }
	    
// 	    cout<<"Entering Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
// 	    DumpIntoNetworkLog(current);
// 	    cout<<"Exiting Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
	     
		bodyrecv += n;
        }
        length += bodylen;
    }

    /* here we connect to the server, gateway, or the forwarder */
    csp->sfd = direct_connect (gw, http, csp);
    if (csp->sfd < 0) {
        if (errno == EINVAL)
	{
            syslog (LOG_NOTICE, "No Such Domain.");
	    printf("No Such Domain.");
	}
        else
	{
            syslog (LOG_ERR, "%s", safe_strerror (errno));
	    printf("%s", safe_strerror (errno));
	}

        freez (hdr);
        return;
    }

    if (gw->forward_host || !http->ssl) {
        /*
         * write the client's (modified) header to the server (along
         * with anything else that may be in the buffer)
         */
        sent = 0;
        while (sent < length) {
            n = write (csp->sfd, hdr+sent, length-sent);
	   //printf("Bakalim burada ne demis: %s ---burada biter\n",hdr+sent);	
            if (n < 0) {
                syslog (LOG_NOTICE, "Failed to send header data to server");
	        printf("Failed to send header data to server");
                freez(hdr);
                return;
            }
	    sent += n;
        }
    } else {
        /*
         * we're running an SSL tunnel and we're not forwarding, so
         * just send the "connect succeeded" message to the client,
         * flush the rest, and get out of the way.
         */
        if (write (csp->cfd, CSUCCEED, sizeof (CSUCCEED) - 1) < 0) {
            freez (hdr);
            return;
        }
    }

    freez (csp->iob->buf);
    memset(csp->iob, '\0', sizeof(csp->iob));
    /* we're finished with the client's header */
    freez (hdr);
    maxfd = (csp->cfd > csp->sfd) ? csp->cfd : csp->sfd;

    /*
     * Pass data between the client and server until one or the other
     * shuts down the connection.
     *
     * FWD: This is quite a violation of the HTTP-1.1
     * standard; it does not properly implement presistent connections.
     * This is no problem however when using netscape as client, but e.g. links
     * causes troubles (because it does the correct thing ;). Doing the right
     * thing is looking for multiple client requests on a single connection
     * that may be to multiple servers. The quick, dirty and easy but wrong way
     * implemented now is to shutdown the connection for reads after the first
     * request has been read. Brrr...
     */

    server_body = 0;

    for (;;) {
        FD_ZERO (&rfds);
        FD_SET (csp->cfd, &rfds);
        FD_SET (csp->sfd, &rfds);
        timeout.tv_usec = 0;
        if (potential_add)
                timeout.tv_sec = 5*tfactor;
        else
                timeout.tv_sec = 60*tfactor;

        n = select (maxfd + 1, &rfds, NULL, NULL, &timeout);

        if (!n) {
            syslog (LOG_NOTICE, "Timed out after %d seconds", timeout.tv_sec);
	    printf("Timed out after %d seconds", timeout.tv_sec);
            return;
        }
        if (n < 0) {
            syslog (LOG_ERR, "%s", safe_strerror (errno));
	    printf("%s", safe_strerror (errno));
            return;
        }

        /*
         * this is the body of the browser's request just read it and
         * write it.
         */
        if (FD_ISSET (csp->cfd, &rfds)) {
            n = read (csp->cfd, buf, sizeof (buf));
            if (n <= 0)
                break;

	    /**Time Lookup Should be Computed at Each read/write on cfd */
	    time(&ntime);
	    current=localtime(&ntime);
	
	
// 	    cout<<"Buffer Read From cfd(3): "<<buf<<endl<<"finito"<<endl;
/*    	Globals::packetParser.packet.sourceIP=strdup(sourceIP);
	Globals::packetParser.packet.destinationIP=strdup(destinationIP);
	Globals::packetParser.packet.accessedURL=strdup(accessedURL);
        Globals::packetParser.packet.packetSize=n;
	
	pthread_mutex_unlock(&(Globals::mutexPacket));
	
	cout<<"A packet(3): "<<endl<<"Accessed URI: "<<Globals::packetParser.packet.accessedURL<<endl;
	cout<<"Source IP: "<<Globals::packetParser.packet.sourceIP<<endl;
	cout<<"Destination IP: "<<Globals::packetParser.packet.destinationIP<<endl;
	cout<<"packetSize: "<<Globals::packetParser.packet.packetSize<<endl;*/
	
// 	cout<<"Entering Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;
// 	DumpIntoNetworkLog(current);
// 	cout<<"Exiting Dump: "<<Globals::packetParser.networkTrafficLog.size()<<endl;

	    packet.packetSize+=n;
	    
	    packet.lastPacketTime.day=current->tm_mday;
	    packet.lastPacketTime.month=current->tm_mon+1;
	    packet.lastPacketTime.year=current->tm_year+1900;
	    packet.lastPacketTime.hour=current->tm_hour;
	    packet.lastPacketTime.minute=current->tm_min;
	    packet.lastPacketTime.second=current->tm_sec;
	
	    //cout<<"Last: "<<packet.lastPacketTime.hour<<":"<<packet.lastPacketTime.minute<<":"<<packet.lastPacketTime.second<<endl;
	    //cout<<"first: "<<packet.firstPacketTime.hour<<":"<<packet.firstPacketTime.minute<<":"<<packet.firstPacketTime.second<<endl;
	    firstInMinute=packet.firstPacketTime.hour*60+packet.firstPacketTime.minute+packet.firstPacketTime.second/60;
	    lastInMinute=packet.lastPacketTime.hour*60+packet.lastPacketTime.minute+packet.lastPacketTime.second/60;			
	    //cout<<"Bura4: "<<lastInMinute-firstInMinute<<endl;
	    
	    if(lastInMinute-firstInMinute>=timePassed)	/**Write to the Database after a minute passes */
	    {
		    DumpIntoNetworkLog(packet);
 		    packet.packetSize=0;
		    timePassed+=5;
	    }
	    //cout<<"Last PAcket Time (4)"<<endl;
	    //cout<<"Packet New Size(3): "<<packet.packetSize<<endl;
	    		
	char *pbuf=buf;
	acceptEncodingFlag=false;
	parseAcceptEncoding(&pbuf,&acceptEncodingFlag);
	//cout<<endl<<endl<<"this is my pbuf: "<<pbuf<<endl;
	if(acceptEncodingFlag) {
		for(copy=0;copy<BUFSIZ && *pbuf ; copy++)
			buf[copy]=pbuf[copy];
		buf[copy]='\0';	
	}
		
	    if (write (csp->sfd, buf, n) != n) {
                syslog (LOG_NOTICE, "Write to: %s failed", http->host);
		printf("Write to: %s failed", http->host);
                return;
            }
	   //printf("Browsers request payload: %s--Finito\n",buf);	

            /* Arf, depending on your browser you may need this */
            if (0 && !http->ssl)
                shutdown (csp->cfd, SHUT_RD);

            continue;
        }

        /*
         * the server wants to talk. it could be the header or the
         * body. if `hdr' is null, then it's the header otherwise
         * it's the body
         */
        if (FD_ISSET (csp->sfd, &rfds)) {
            n = read (csp->sfd, buf, sizeof (buf));
// 	    cout<<"This is also read "<<buf<<endl;
	    if (n < 0) {
                syslog (LOG_NOTICE, "Read from: %s failed: ", http->host);
		printf("Read from: %s failed: ", http->host);
                syslog (LOG_ERR, "%s", safe_strerror (errno));
		printf("%s", safe_strerror (errno));
                return;
            }
            if (!n)
                break;                /* "game over, man" */

	    
	    

		/** Filtering Content Code */
 		
	    
	    	filter.setText(buf);
                doIFilter=false;
 		doIFilter=filter.filterContent(); 
		//cout<<endl<<"Here I am: "<<buf<<endl<<endl;

            /*
             * if this is an SSL connection or we're in the body
             * of the server document, just write it to the
             * client.
             */
            //printf("Here is the packet you write to cfd: %s---Bitis\n",buf);
		
// 		cout<<"Burada cfd basacaydik "<<endl;		
        /**Burada Serverdan okunan sey body */    
	if (server_body || http->ssl) {
                /* just write */
                
  		if(!doIFilter){
// 		    cout<<"Buffer sfd(1): "<<buf<<endl;   

         
              		if (write (csp->cfd, buf, n) != n) {
	                     syslog (LOG_NOTICE, "Write to client failed");
	 		    printf("Write to client failed");
	                    syslog (LOG_ERR, "%s", safe_strerror (errno));
			    printf("%s", safe_strerror (errno));
	                    return;
	                }
			time(&ntime);
			current=localtime(&ntime);
		
			packet.packetSize+=n;
			
			packet.lastPacketTime.day=current->tm_mday;
			packet.lastPacketTime.month=current->tm_mon+1;
			packet.lastPacketTime.year=current->tm_year+1900;
			packet.lastPacketTime.hour=current->tm_hour;
			packet.lastPacketTime.minute=current->tm_min;
			packet.lastPacketTime.second=current->tm_sec;
	
			//cout<<"Last: "<<packet.lastPacketTime.hour<<":"<<packet.lastPacketTime.minute<<":"<<packet.lastPacketTime.second<<endl;
			//cout<<"first: "<<packet.firstPacketTime.hour<<":"<<packet.firstPacketTime.minute<<":"<<packet.firstPacketTime.second<<endl;
		
			firstInMinute=packet.firstPacketTime.hour*60+packet.firstPacketTime.minute+packet.firstPacketTime.second/60;
			lastInMinute=packet.lastPacketTime.hour*60+packet.lastPacketTime.minute+packet.lastPacketTime.second/60;
			
			//if(firstTime2){
			//	DumpIntoNetworkLog(packet);
                          //      packet.packetSize=0;		
			//	firstTime2=false;
			//}
			//cout<<"Bura5: "<<lastInMinute-firstInMinute<<endl;
	    	
			if(lastInMinute-firstInMinute>=timePassed)	/**Write to the Database after a minute passes */
			{
				DumpIntoNetworkLog(packet);
                                packet.packetSize=0;
				timePassed+=5;
			}
			//cout<<"Last PAcket Time (5)"<<endl;			
			//cout<<"Packets New Size:(write To cfd) "<<packet.packetSize<<endl;
// 			cout<<"Write to cfd(0): "<<buf<<endl<<"finito"<<endl;
                  } 
                 else {
			
			
	//		parseBody(buf,sendNetCheck);
	//cout<<"SendNetCheck: "<<sendNetCheck<<endl;
	//		sendNetCheck=sendNetCheck+stringBlock;
			if (write (csp->cfd,CSUCCEED,sizeof(CSUCCEED)-1)<0 ) {
	                     syslog (LOG_NOTICE, "Write to client failed");
	 		    printf("Write to client failed");
	                    syslog (LOG_ERR, "%s", safe_strerror (errno));
			    printf("%s", safe_strerror (errno));
	                    return;
	                }
			DumpIntoNetworkLog(packet);	
			DumpIntoFilteredContent(packet);
			return;

		}  
                continue;
            } else {
                /*
                 * we're still looking for the end of the
                 * server's header ...
                 */

                add_to_iob (csp, buf, n);
                while ((p = get_header (csp))) {
                    /* see following note */
                    if (*p == '\0')
                        break;
                    enlist (csp->headers, p);
                    freez (p);
                }

                /*
                 * NOTE: there are no "empty" headers so if
                 * the pointer `p' is not NULL we must assume
                 * that we reached the end of the buffer
                 * before we hit the end of the header.
                 *
                 * Since we have to wait for more from the
                 * server before we can parse the headers we
                 * just continue here.
                 */

                if (p)
                    continue;

                /*
                 * add the blank line at the end of the
                 * header
                 */
                enlist (csp->headers, "");
                hdr = list_to_text (csp->headers);
                n = strlen (hdr);

                /*
                 * write the server's (modified) header to
                 * the client (along with anything else that
                 * may be in the buffer)
                 */
		//printf("this is the second time: %s--the end\n",hdr);
// 		if(n)
// 		cout<<"Buffer sfd(2): "<<hdr<<endl;
            

                if ((n && write (csp->cfd, hdr, n) != n)
                    || (flush_socket (csp->cfd, csp) < 0)) {
                    /*
                     * the write failed, so don't bother
                     * mentioning it to the client... it
                     * probably can't hear us anyway.
                     */
                    freez (hdr);
                    return;
                }
		time(&ntime);
		current=localtime(&ntime);
		
		packet.packetSize+=n;
		
		packet.lastPacketTime.day=current->tm_mday;
		packet.lastPacketTime.month=current->tm_mon+1;
		packet.lastPacketTime.year=current->tm_year+1900;
		packet.lastPacketTime.hour=current->tm_hour;
		packet.lastPacketTime.minute=current->tm_min;
		packet.lastPacketTime.second=current->tm_sec;
	
		firstInMinute=packet.firstPacketTime.hour*60+packet.firstPacketTime.minute+packet.firstPacketTime.second/60;
		lastInMinute=packet.lastPacketTime.hour*60+packet.lastPacketTime.minute+packet.lastPacketTime.second/60;			
		//cout<<"BuraD: "<<lastInMinute-firstInMinute<<endl;
	    
		if(lastInMinute-firstInMinute>=timePassed)	/**Write to the Database after a minute passes */
		{
			DumpIntoNetworkLog(packet);
			packet.packetSize=0;
			timePassed+=5;
		}
		//cout<<"Last PAcket Time (6)"<<endl;
		//cout<<"Last: "<<packet.lastPacketTime.hour<<":"<<packet.lastPacketTime.minute<<":"<<packet.lastPacketTime.second<<endl;
		//cout<<"first: "<<packet.firstPacketTime.hour<<":"<<packet.firstPacketTime.minute<<":"<<packet.firstPacketTime.second<<endl;
		//cout<<"Packets New Size:(write cfd Hdr) "<<packet.packetSize<<endl;
// 		cout<<"Write to cfd(1): "<<hdr<<endl<<"finito"<<endl;
                /* we're finished with the server's header */

                freez (hdr);
                server_body = 1;
            }

	    firstInMinute=packet.firstPacketTime.hour*60+packet.firstPacketTime.minute+packet.firstPacketTime.second/60;
	    lastInMinute=packet.lastPacketTime.hour*60+packet.lastPacketTime.minute+packet.lastPacketTime.second/60;			
	    //cout<<"BuraA: "<<lastInMinute-firstInMinute<<endl;
	    
	    if(lastInMinute-firstInMinute>=timePassed)	/**Write to the Database after a minute passes */
	    {
		    DumpIntoNetworkLog(packet);
		    packet.packetSize=0;  
                    timePassed+=5;
	    }
	    continue;
	    
        }
        return;                        /* huh? we should never get here */
    }
}

void Proxy::serve (struct client_state *csp,char *sourceIP) {

    chat (csp,sourceIP);
    shutdown (csp->cfd, SHUT_RDWR);
    close (csp->cfd);

    if (csp->sfd >= 0)
        close (csp->sfd);
}

void Proxy::execute (void *p) {
    char buf[BUFSIZ];
    int bfd[MAXADDRESSES + 1];
    char *addr;
    int i;

 

    configfile = default_configfile;
/*    for (i = 1; i < argc; i++) {
        if (!strcmp (argv[i], "-d") )
            debug = 1;
        else if (!strcmp (argv[i], "-c"))
            configfile = strdup(argv[++i]);
        else if (!strcmp (argv[i], "-l")) {
                int j;

                i++;
                j = 0;
                while (facilitynames[j].c_name) {
                    if (!strcmp(argv[i], facilitynames[j].c_name))
                        logfacility = facilitynames[j].c_val;
                    j++;
                }
        }
        else if (!strcmp (argv[i], "-q"))
                quiet = 1;
        else if (!strcmp (argv[i], "-v"))
                quiet = -1;
        else if (!strcmp (argv[i], "-h")) {
                fprintf (stdout, "www6to4 -q -v -h  -d -c <configfile> -l <syslogfacility>\n");
                fprintf (stdout, "-q: quiet (less logging)\n");
                fprintf (stdout, "-v: verbose (more logging)\n");
                fprintf (stdout, "-d: debug mode (do not fork)\n");
        }
    } */

    if (debug)
        openlog("www6to4", LOG_PID | LOG_NDELAY | LOG_CONS | LOG_PERROR, logfacility);
    else
        openlog("www6to4", LOG_PID | LOG_NDELAY | LOG_CONS, logfacility);

    syslog (LOG_NOTICE, "%s", VERSION);
    if (quiet > 0)
            setlogmask (LOG_UPTO (LOG_ERR));
    else if (!quiet)
        setlogmask (LOG_UPTO (LOG_NOTICE));
    else if (quiet < 0)
        setlogmask (LOG_UPTO (LOG_INFO));

    if (configfile) {
        FILE *configfp = fopen (configfile, "r");

        if (!configfp) {
            syslog (LOG_ERR, "can't open configuration file '%s': ", configfile);
	    printf("can't open configuration file '%s': ", configfile);
            syslog (LOG_ERR, "%s", safe_strerror (errno));
	    printf("%s", safe_strerror (errno));
            exit (1);
        } else {
            int line_num = 0;

            while (fgets (buf, sizeof (buf), configfp)) {
                char cmd[BUFSIZ], arg[BUFSIZ], tmp[BUFSIZ];
                    char *p, *q;

                line_num++;

                strcpy (tmp, buf);
                p = strpbrk (tmp, "#\r\n");
                if (p)
                    *p = '\0';
                p = tmp;

                /* leading skip whitespace */
                while (*p && ((*p == ' ') || (*p == '\t')))
                    p++;

                q = cmd;
                while (*p && (*p != ' ') && (*p != '\t'))
                    *q++ = *p++;
                *q = '\0';

                while (*p && ((*p == ' ') || (*p == '\t')))
                    p++;

                strcpy (arg, p);
                p = arg + strlen (arg) - 1;

                /* ignore trailing whitespace */
                while (*p && ((*p == ' ') || (*p == '\t')))
                    *p-- = '\0';

                if (*cmd == '\0')
                    continue;

                /* insure the command field is lower case */
                for (p = cmd; *p; p++)
                    if (isupper (*p))
                        *p = tolower (*p);

                if (!strcmp (cmd, "listen-to")) {
                    strncpy (hostname, arg, MAXHOST);
                    continue;
                }
                if (!strcmp (cmd, "pattern")) {
                    if (++lastpattern >= MAXPATTERNS)
                        syslog (LOG_ERR, "Too many patterns, more than %d.", MAXPATTERNS);
                    else
                        patterns[lastpattern] = strdup(arg);
                    continue;
                }
                if (!strcmp (cmd, "listen-port")) {
                    strncpy (hostport, arg, MAXPORT);
                    continue;
                }
                if (!strcmp (cmd, "forwardfile")) {
                    forwardfile = strdup (arg);
                    continue;
                }
                if (!strcmp (cmd, "timeout-factor")) {
                    tfactor = atoi (arg);
                    if (tfactor < 0 || tfactor > 10) {
                        tfactor = 1;
                        syslog (LOG_ERR, "Illegal Timeout factor.");
                    }
                    continue;
                }
                syslog (LOG_ERR,
                         "Unrecognized directive in configuration file "
                         "at line number %d", line_num);
                exit(1);
            }
            fclose (configfp);
        }
    }

    if (forwardfile)
        add_loader ((int (*)())(load_forwardfile));

    addr = strtok (hostname, ",");
    if (!strcmp (addr, "*")) {
        bfd[0] = open_server_socket (NULL, hostport, AF_INET);
        bfd[1] = open_server_socket (NULL, hostport, AF_INET6);
        bfd[2] = 0;
    } else {
        for (i = 0; i < MAXADDRESSES; i++) {
            bfd[i] = open_server_socket (addr, hostport, AF_UNSPEC);
            addr = strtok (NULL, ",");
            if (!addr)
                break;
        }
        bfd[i + 1] = 0;
    }

    for (i = 0; (i < MAXADDRESSES) && bfd[i]; i++) {
        if (bfd[i] < 0) {
            syslog (LOG_ERR, "Can't bind %s:%s: ",
                     hostname ? hostname : "INADDR_ANY", hostport);
	   printf("Can't bind %s:%s: ",
                     hostname ? hostname : "INADDR_ANY", hostport);
            syslog (LOG_ERR, "%s", safe_strerror (errno));
	   printf("%s", safe_strerror (errno));
            syslog (LOG_ERR, "There may be another www6to4 or some other "
                     "proxy running on port %s", hostport);
		printf("There may be another www6to4 or some other "
                     "proxy running on port %s", hostport);
            exit (1);
        }
    }

    signal (SIGPIPE, SIG_IGN);
    signal (SIGCHLD, SIG_IGN);

    for (;;) {
        int cfd;
        struct client_state *csp;

#ifdef DMALLOC
        dmalloc_log_unfreed();
#endif

        /* cleanup zombie children */
        while (waitpid (-1, NULL, WNOHANG) > 0) {}

        cfd = accept_conn (bfd);
        if (cfd < 0)
            continue;

        csp = (struct client_state *) calloc (1, sizeof (*csp));
        if (!csp) {
            syslog (LOG_ERR, "malloc(%d) for csp failed: ", sizeof (*csp));
	    printf("malloc(%d) for csp failed: ", sizeof (*csp));
            syslog (LOG_ERR, "%s", safe_strerror (errno));
	    printf("%s", safe_strerror (errno));
            close (cfd);
            sleep (3);
            continue;
        }

        csp->cfd = cfd;
        csp->sfd = -1;

        if (run_loader (csp)) {
            syslog (LOG_ERR, "A loader failed - must exit");
	   printf("A loader failed - must exit");
            exit (1);
        }

        if (!debug) {
            int child_id;
            child_id = fork ();

            if (child_id < 0) {        /* failed */
                syslog (LOG_ERR, "Can't fork");
	    	printf("Can't fork");
                syslog (LOG_ERR, "%s", safe_strerror (errno));
		printf("%s", safe_strerror (errno));
                sprintf (buf, "%s: can't fork: errno = %d", VERSION, errno);
		printf("%s: can't fork: errno = %d", VERSION, errno);
                write (csp->cfd, buf, strlen (buf));
                close (csp->cfd);
                sleep (5);
                continue;
            }

            if (!child_id) { /* child */
                    if (setpgid (0, getpid ()) == -1) {
                    syslog (LOG_ERR, "setpgid() failed with %d", errno);
		    printf("setpgid() failed with %d", errno);
                    syslog (LOG_ERR, "%s", safe_strerror (errno));
		    printf("%s", safe_strerror (errno));
                    exit(1);
                }
                serve (csp,sourceIP);
                freez (csp);
                exit (0);
            } else {         /* parent */
                close (csp->cfd);
                freez (csp);
            }
        } else {
            serve (csp,sourceIP);
            freez (csp);
        }
    }
    /* NOTREACHED */
}


char *safe_strerror (int err) {
    char buf[BUFSIZ];
    char *s = NULL;

#ifndef   NOSTRERROR
    s = strerror (err);
#endif                                /* NOSTRERROR */

    if (!s) {
        sprintf (buf, "(errno = %d)", err);
        s = buf;
    }
    return s;
}

int Proxy::flush_socket (int fd, struct client_state * csp) {
    struct iob *iob = csp->iob;
    int n = iob->eod - iob->cur;

    if (n <= 0)
        return (0);

    n = write (fd, iob->cur, n);
    iob->eod = iob->cur = iob->buf;

    return (n);
}








/** loaders */
int loaders_strncmpic(const char *s1, const char *s2, size_t n)
{
   if (n <= 0) return(0);
   if (!s1) s1 = "";
   if (!s2) s2 = "";
   
   while (*s1 && *s2)
   {
      if ( ( *s1 != *s2 ) && ( ijb_tolower(*s1) != ijb_tolower(*s2) ) )
      {
         break;
      }

      if (--n <= 0) break;

      s1++, s2++;
   }
   return(ijb_tolower(*s1) - ijb_tolower(*s2));

}

void unload_forwardfile (struct forward_spec *b) {

    if (!b)
        return;

    unload_forwardfile (b->next);

    if (b->url) {
        struct url_spec *url = b->url;

        freez (url->spec);
        freez (url->domain);
        freez (url->dbuf);
        freez (url->dvec);
        freez (url->path);
    }

    freez (b->gw->forward_host);
    freez (b);
}

int load_forwardfile (struct client_state *csp) {
    static struct stat prev[1], curr[1];
    struct forward_spec *b, *bl;
    char buf[BUFSIZ], *p, *q, *tmp;
    char *vec[2], *port;
    int n;
    struct file_list *fs;
    struct url_spec url[1];
    FILE *fp;

    if (stat (forwardfile, curr) < 0)
        goto load_forwardfile_error;

    if (current_forwardfile && (prev->st_mtime == curr->st_mtime)) {
        /* ok, the forwardfile hasn't changed a bit */
        csp->flist = current_forwardfile;
        return (0);
    }

    fs = (struct file_list *) calloc (1, sizeof (*fs));
    bl = (struct forward_spec *) calloc (1, sizeof (*bl));
    if (!fs || !bl)
        goto load_forwardfile_error;

    fs->f = bl;
    *prev = *curr;
    if (csp)
        csp->flist = fs;

    fp = fopen (forwardfile, "r");
    if (!fp)
        goto load_forwardfile_error;

    tmp = NULL;
    while (fgets (buf, sizeof (buf), fp)) {

        freez (tmp);

        p = strpbrk (buf, "\r\n");
        if (p)
            *p = '\0';

        /* comments */
        p = strchr (buf, '#');
        if (p)
            *p = '\0';
        /* skip blank lines */
        if (*buf == '\0')
            continue;

        tmp = strdup (buf);
        n = ssplit (tmp, " \t", vec, SZ (vec), 1, 1);
        if (n != 2) {
            syslog (LOG_ERR, "error in forwardfile: %s\n", buf);
            continue;
        }

        strcpy (buf, vec[0]);

        /* skip blank lines */
        if (*buf == '\0')
            continue;

        /* allocate a new node */
        b = (forward_spec *)calloc (1, sizeof (*b));
        if (!b) {
            fclose (fp);
            goto load_forwardfile_error;
        }
        /* add it to the list */
        b->next = bl->next;
        bl->next = b;

        p = strstr (buf, "://");
        if (p) {
            if (!loaders_strncmpic (buf, "http://", 7)) {
                b->url->proto = HTTP_PROTO;
                memmove(buf, buf+7, BUFSIZ-7);
            } else if (!loaders_strncmpic (buf, "https://", 8)) {
                b->url->proto = HTTPS_PROTO;
                memmove(buf, buf+8, BUFSIZ-8);
            } else if (!loaders_strncmpic (buf, "ftp://", 6)) {
                /* we do not support ftp proxying, but we may forward */
                b->url->proto = FTP_PROTO;
                memmove(buf, buf+6, BUFSIZ-6);
            } else {
                b->url->proto = UNDEF_PROTO;
            }
        }

        p = strchr (buf, '/');
        if (p) {
            b->url->path = strdup (p);
            b->url->pathlen = strlen (b->url->path);
            *p = '\0';
        } else {
            b->url->path = NULL;
            b->url->pathlen = 0;
        }

        p = strchr (buf, ':');
        if (!p)
            port = "0";
        else {
            *p++ = '\0';
            port = strdup (p);
        }

        b->url->port = port;
        b->url->domain = strdup (buf);
        if (!b->url->domain) {
            fclose (fp);
            goto load_forwardfile_error;
        }
        /* split domain into components */
        *url = dsplit (b->url->domain);
        b->url->dbuf = url->dbuf;
        b->url->dcnt = url->dcnt;
        b->url->dvec = url->dvec;

        /* now parse the forwarding spec */
        p = vec[1];

        if (strcmp (p, ".") != 0) {
            b->gw->forward_host = strdup (p);

            p = strchr (b->gw->forward_host, ':');
            if (p) {
                *p++ = '\0';
                b->gw->forward_port = strdup (p);
            }
            if (b->gw->forward_port <= 0)
                b->gw->forward_port = "8000";
        }
    }

    freez (tmp);
    fclose (fp);

    /* the old one is now obsolete */
    if (current_forwardfile)
        current_forwardfile->unloader =(void (*) ())(unload_forwardfile);

    current_forwardfile = fs;
    return (0);

load_forwardfile_error:
    syslog(LOG_ERR, "Can't load forwardfile '%s': ", forwardfile);
    syslog (LOG_ERR, "%s", safe_strerror (errno));
    return (-1);
}


/*
 * strsav() takes a pointer to a string stored in a dynamically allocated
 * buffer and a pointer to a string and returns a pointer to a new
 * dynamically allocated space that contains the concatenation of the two
 * input strings the previous space is free()'d by realloc().
 */
char *strsav (char *old, char *text_to_append) {
    int old_len, new_len;
    char *p;

    if (!text_to_append || (*text_to_append == '\0'))
        return (old);

    if (old)
        old_len = strlen (old);
    else
        old_len = 0;

    new_len = old_len + strlen (text_to_append) + 1;

    if (old) {
        p = (char *)realloc (old, new_len);
        if (!p) {
            syslog(LOG_ERR, "Realloc(%d) bytes failed!\n", new_len);
            exit (1);
        }
    }
    else {
        p = (char *)malloc (new_len);
        if (!p) {
            syslog(LOG_ERR, "Realloc(%d) bytes failed!\n", new_len);
            exit (1);
        }
    }

    strcpy (p + old_len, text_to_append);
    return (p);
}

void add_loader (int (*loader) ()) {
    int i;

    for (i = 0; i < NLOADERS; i++) {
        if (!loaders[i]) {
            loaders[i] = (int (*) (client_state *))loader;
            break;
        }
    }
}

int run_loader (struct client_state *csp) {
    int ret = 0;
    int i;

    for (i = 0; i < NLOADERS; i++) {
        if (!loaders[i])
            break;

        ret |= (loaders[i]) (csp);
    }
    return (ret);
}

/** parsers */

/* case insensitive string comparison */
int strcmpic (char *s1, char *s2) {
    while (*s1 && *s2) {
        if ((*s1 != *s2)
            && (tolower (*s1) != tolower (*s2))) {
            break;
        }
        s1++, s2++;
    }
    return (tolower (*s1) - tolower (*s2));
}

int strncmpic (char *s1, char *s2, size_t n) {
    if (n <= 0)
        return (0);

    while (*s1 && *s2) {
        if ((*s1 != *s2)
            && (tolower (*s1) != tolower (*s2))) {
            break;
        }
        if (--n <= 0)
            break;

        s1++, s2++;
    }
    return (tolower (*s1) - tolower (*s2));
}

/*
 * Append buf to the current csp->iob reallocating it if necessary.
 */
int add_to_iob (struct client_state *csp, char *buf, int n) {
    struct iob *iob = csp->iob;
    int have, need;
    char *p;
    have = iob->eod - iob->cur;

    if (n <= 0)
        return (have);

    need = have + n;

    p = (char *)malloc (need + 1);
    if (!p) {
        syslog (LOG_ERR, "Malloc() iob failed");
	syslog (LOG_ERR, "%s", safe_strerror (errno));
        return (-1);
    }

    if (have) {
        /* there is something in the buffer - save it */
        memcpy (p, iob->cur, have);

        /* replace the buffer with the new space */
        freez (iob->buf);
        iob->buf = p;

        /* point to the end of the data */
        p += have;
    } else {
        /* the buffer is empty, free it and reinitialize */
        freez (iob->buf);
        iob->buf = p;
    }

    /* copy the new data into the iob buffer */
    memcpy (p, buf, n);

    /* point to the end of the data */
    p += n;

    /* null terminate == cheap insurance */
    *p = '\0';

    /* set the pointers to the new values */
    iob->cur = iob->buf;
    iob->eod = p;

    return (need);
}

/*
 * this (odd) routine will parse the csp->iob and return one of the
 * following:
 *
 * 1) a pointer to a dynamically allocated string that contains a
 * header line 2) NULL indicating that the end of the header was
 * reached 3) "" indicating that the end of the iob was reached before
 * finding a complete header line.  
 */
char *get_header (struct client_state *csp) {
    struct iob *iob = csp->iob;
    char *p, *q, *ret;

    /* If we don't have it we don't have it */
    if (!iob->cur)
        return ("");
    p = strchr (iob->cur, '\n');
    if (!p)
        return ("");                /* couldn't find a complete header */

    /* Null terminate the header line and duplicate it */
    *p = '\0';
    ret = strdup (iob->cur);
    if ((q = strchr (ret, '\r'))) /* Snip that annoyin character... */
        *q = '\0';

    /* Move the start of the buffer to point at the next header data. */
    iob->cur = p + 1;

    /* Is this a blank linke (i.e. the end of the header) ? */
    if (*ret == '\0') {
        freez (ret);
        return (NULL);
    }
    return (ret);
}

void free_http_request (struct http_request *http) {
    freez (http->cmd);
    freez (http->gpc);
    freez (http->host);
    freez (http->hostport);
    freez (http->path);
    freez (http->ver);
}

/* parse out the host and port from the URL */
void parse_http_request (char *req, struct http_request *http, struct client_state *csp) {
		
    char *buf, *v[10], *url, *p;
    int n;

    
//     cout<<"parse_http_request(req): "<<req<<endl<<"parse_http_request(http): "<<http->cmd<<" "<<http->gpc<<" "<<http->host<<" "<<http->port<<" "<<http->path<<" "<<http->ver<<" "<<http->hostport<<" "<<http->ssl<<" "<<http->proto<<endl;
//     
    memset (http, '\0', sizeof (*http));
    /* Store the request as http->cmd */
    http->cmd = strdup (req);
    /* Copy the request to buf */
    buf = strdup (req);

    n = ssplit (buf, " \r\n", v, SZ (v), 1, 1);
    if (n == 3) {
        /* this could be a CONNECT request */
        if (!strcmpic (v[0], "connect")) {
            http->ssl = 1;
            http->gpc = strdup (v[0]);
            http->hostport = strdup (v[1]);
            http->ver = strdup (v[2]);
        }
        /* or it could be a GET or a POST */
        if (!strcmpic (v[0], "get")
            || !strcmpic (v[0], "head")
            || !strcmpic (v[0], "post")) {

            http->ssl = 0;
            http->gpc = strdup (v[0]);
            url = v[1];
            http->ver = strdup (v[2]);

            if (!strncmpic (url, "http://", 7)) {
                http->proto = HTTP_PROTO;
                url += 7;
            } else if (!strncmpic (url, "https://", 8)) {
                http->proto = HTTPS_PROTO;
                url += 8;
            } else if (!strncmpic (url, "ftp://", 6)) {
                /* we do not support ftp proxying, but we may forward */
                http->proto = FTP_PROTO;
                url += 6;
            } else {
                http->proto = UNDEF_PROTO;
                url = NULL;
            }

            if (url && (p = strchr (url, '/'))) {
                http->path = strdup (p);
                *p = '\0';
                http->hostport = strdup (url);
            }
        }
    }
    freez (buf);

    if (!http->hostport) {
        free_http_request (http);
        return;
    }

    buf = strdup (http->hostport);
    n = ssplit (buf, ":", v, SZ (v), 1, 1);
    if (n == 1) {
        http->host = strdup (v[0]);
        http->port = "80";
    }
    if (n == 2) {
        http->host = strdup (v[0]);
        http->port = strdup (v[1]);
    }
    freez (buf);

    if (!http->host)
        free_http_request (http);

    if (!http->path)
        http->path = strdup ("");
}

/*
 * ssplit() - split a string (in-place) into fields s = string to split c =
 * list of characters to be used as field separators (if NULL, use default
 * separators of space, tab, and newline)
 *
 * v = vector into which field pointers are placed n = number of fields in
 * vector
 *
 * m = flag indicating whether to treat strings of field separators as
 * indicating multiple fields
 *
 * l = flag indicating whether to ignore leading field separators
 */

int ssplit (char *s, char *c, char *v[], int n, int m, int l) {
    char t[256];
    char **x = NULL;
    int xsize = 0;
    unsigned char *p, b;
    int xi = 0;
    int vi = 0;
    int i;
    int last_was_null;
    if (!s)
        return (-1);

    memset (t, '\0', sizeof (t));

    p = (unsigned char *) c;

    if (!p)
        p = (unsigned char *) " \t";        /* default field separators */

    while (*p)
        t[*p++] = 1;                /* separator  */

    t['\0'] = 2;                /* terminator */
    t['\n'] = 2;                /* terminator */

    p = (unsigned char *) s;

    if (l) {                        /* are we to skip leading separators ? */
        while ((b = t[*p]) != 2) {
            if (b != 1)
                break;
            p++;
        }
    }
    xsize = 256;

    x = (char **) calloc (xsize, sizeof (char *));
    x[xi++] = (char *) p;        /* first pointer is the beginning of string */

    /* first pass:  save pointers to the field separators */
    while ((b = t[*p]) != 2) {
        if (b == 1) {                /* if the char is a separator ... */
            *p++ = '\0';        /* null terminate the substring */

            if (xi == xsize) {
                /* get another chunk */
                int new_xsize = xsize + 256;
                char **new_x = (char **)
                calloc (new_xsize, sizeof (char *));
                for (i = 0; i < xsize; i++)
                    new_x[i] = x[i];

                free (x);
                xsize = new_xsize;
                x = new_x;
            }
            x[xi++] = (char *) p;        /* save pointer to beginning of next
                                         * string */
        } else
            p++;
    }
    *p = '\0';                        /* null terminate the substring */

    /* second pass: copy the relevant pointers to the output vector */
    last_was_null = 0;
    for (i = 0; i < xi; i++) {
        if (m) {
            /* there are NO null fields */
            if (*x[i] == 0)
                continue;
        }
        if (vi < n)
            v[vi++] = x[i];
        else {
            free (x);
            return (-1);        /* overflow */
        }
    }
    free (x);

    return (vi);
}


struct gateway *forward_url (struct http_request *http, struct client_state *csp) {
    struct file_list *fl;
    struct forward_spec *b;
    struct url_spec url[1];

    fl = csp->flist;
    if (!fl)
        return (&gw_default);

    b = (forward_spec *)fl->f;
    if (!b)
        return (&gw_default);

    *url = dsplit (http->host);

    /* if splitting the domain fails, punt */
    if (!url->dbuf)
        return (&gw_default);

    for (b = b->next; b; b = b->next) {
        if (!b->url->proto || b->url->proto == http->proto) {
            if (!atoi (b->url->port) || atoi(b->url->port) == atoi(http->port)) {
                if ((b->url->domain[0] == '\0') || !domaincmp(b->url, url)) {
                    if (!b->url->path || !strncmp (b->url->path, http->path, b->url->pathlen)) {
                        freez (url->dbuf);
                        freez (url->dvec);
                        return (b->gw);
                     }
                }
            }
        }
    }
    freez (url->dbuf);
    freez (url->dvec);
    return (&gw_default);
}

/*
 * dsplit() takes a domain and returns a pointer to a url_spec structure
 * populated with dbuf, dcnt and dvec.  the other fields in the structure
 * that is returned are zero.
 *
 */
struct url_spec dsplit (char *domain) {
    struct url_spec ret[1];
    char *v[BUFSIZ];
    int size;
    char *p;
    memset (ret, '\0', sizeof (*ret));

    if ((p = strrchr (domain, '.'))) {
        if (*(++p) == '\0')
            ret->toplevel = 1;
    }
    ret->dbuf = strdup (domain);

    /* map to lower case */
    for (p = ret->dbuf; *p; p++)
        *p = tolower (*p);

    /* split the domain name into components */
    ret->dcnt = ssplit (ret->dbuf, ".", v, SZ (v), 1, 1);

    if (ret->dcnt <= 0) {
        memset (ret, '\0', sizeof (ret));
        return (*ret);
    }
    /* save a copy of the pointers in dvec */
    size = ret->dcnt * sizeof (*ret->dvec);

    ret->dvec = (char **)malloc (size);
    if (ret->dvec)
        memcpy (ret->dvec, v, size);

    return (*ret);
}

/*
 * the "pattern" is a domain that may contain a '*' as a wildcard. the "fqdn"
 * is the domain name against which the patterns are compared.
 *
 * domaincmp("a.b.c" , "a.b.c") => 0 (MATCH) 
 * domaincmp("a*.b.c", "a.b.c") => 0 (MATCH)
 * domaincmp("b.c"   , "a.b.c") => 0 (MATCH)
 * domaincmp(""      , "a.b.c") => 0 (MATCH)
 */
int domaincmp (struct url_spec * pattern, struct url_spec * fqdn) {
    char **pv, **fv;    /* vectors  */
    int pn, fn;         /* counters */
    char *p, *f;        /* chars    */
    pv = pattern->dvec;
    pn = pattern->dcnt;

    fv = fqdn->dvec;
    fn = fqdn->dcnt;

    while ((pn > 0) && (fn > 0)) {
        p = pv[--pn];
        f = fv[--fn];

        while (*p && *f && (*p == tolower (*f)))
            p++, f++;

        if ((*p != tolower (*f)) && (*p != '*'))
            return (1);
    }

    if (pn > 0)
        return (1);

    return (0);
}

/*
 * Parse content-length header
 */
int parse_content_length(char *p) {
    char *number;
    long length;

    number = strrchr(p, ':');
    if (!number) /* No ':'? Return 0 */
        return 0;

    number++;
    length = strtol(number, (char **)NULL, 0);

    return (int) length;
}

/*
 * h = pointer to list 'dummy' header
 * s = string to add to the list
 */
void enlist (struct list * h, char *s) {
    struct list *n = (struct list *) malloc (sizeof (*n));
    struct list *l;
    if (n) {
        n->str = strdup (s);
        n->next = NULL;

        if ((l = h->last)) {
            l->next = n;
        }
        else {
            h->next = n;
        }

        h->last = n;
    }
}

void destroy_list (struct list * h) {
    struct list *p, *n;
    for (p = h->next; p; p = n) {

        n = p->next;

        freez (p->str);

        freez (p);
    }

    memset (h, '\0', sizeof (*h));
}

char *list_to_text (struct list * h) {
    struct list *p;
    char *ret = NULL;
    char *s;
    int size;
    size = 0;

    for (p = h->next; p; p = p->next) {
        if (p->str) {
            size += strlen (p->str) + 2;
        }
    }

    ret =(char *) malloc (size + 1);
    if (!ret)
        return (NULL);

    ret[size] = '\0';

    s = ret;

    for (p = h->next; p; p = p->next) {
        if (p->str) {
            strcpy (s, p->str);
            s += strlen (s);
            *s++ = '\r';
            *s++ = '\n';
        }
    }

    return (ret);
}

/** socket */

int open_server_socket (char *hostname, char *port, int af) {
    struct addrinfo hints, *r, *res;
    int s = -1, error;
    int reuse_addr = 1;

    memset (&hints, 0, sizeof (hints));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = af;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = 0;

    error = getaddrinfo (hostname, port, &hints, &res);
    if (!error) {
        for (r = res; r; r = r->ai_next) {
            s = socket (r->ai_family, r->ai_socktype, r->ai_protocol);
            if (s < 0)
             { printf("Can not open the socket: %d, %d, %d\n",r->ai_family,r->ai_socktype,r->ai_protocol);
	   	continue;
	     }
	
	printf(" open the socket: %d, %d, %d\n",r->ai_family,r->ai_socktype,r->ai_protocol);

            setsockopt (s, SOL_SOCKET, SO_REUSEADDR, &reuse_addr, sizeof (reuse_addr));
	
            if (bind (s, r->ai_addr, r->ai_addrlen) < 0) {
		close (s);
                s = -1;
                continue;
            }
            if (listen (s, 10) < 0) {
                close (s);
                s = -1;
                continue;
            }
            break;
        }
    }

    if (res)
        freeaddrinfo (res);

    if (s < 0)
        return (-1);

    if (!hostname)
        {syslog (LOG_INFO, "Listening to wildcard address for family %d on port %s.\n", af, port);
	printf("Listening to wildcard address for family %d on port %s.\n", af, port);
	}
    else
	{
        	syslog (LOG_INFO, "Listening to %s for family %d on port %s.\n", hostname, af, port);
		printf("Listening to %s for family %d on port %s.\n", hostname, af, port);
	}
    return s;
}

int accept_conn (int *fds) {
    struct sockaddr_storage ra;
    int fd, afd, ralen, i, last;
    fd_set rfds;
    struct timeval tv[1];
    char remote_ip_str[NI_MAXHOST];
    int clientIPCopy;	

    /* wait for connection to complete */
    FD_ZERO (&rfds);
    i = 0;
    while (fds[i] > 0) {
        FD_SET (fds[i], &rfds);
        i++;
    }
    last = i - 1;

    tv->tv_sec = 10;
    tv->tv_usec = 0;

    if (select (fds[last] + 1, &rfds, NULL, NULL, tv) <= 0)
        return -1;

    i = 0;
    while (fds[i] > 0 && !FD_ISSET (fds[i], &rfds)) {
        i++;
    }

    fd = fds[i];
    if (!fd)
        return -1;

    ralen = sizeof ra;
    do {
        afd = (int)accept (fd, (struct sockaddr *) & ra,(socklen_t *) &ralen);
    } while (afd < 1 && errno == EINTR);

    if (afd < 0)
        return (-1);

    getnameinfo ((struct sockaddr *) & ra, ralen, remote_ip_str, 
                 NI_MAXHOST, NULL, NI_MAXSERV, NI_NUMERICHOST);
    
   sourceIP=strdup(remote_ip_str);
   //cout<<endl<<endl<<"Packet Source IP: "<<sourceIP<<endl;
	syslog (LOG_INFO, "Accepted client from: %s.\n", remote_ip_str);
   //printf("Accepted client from: %s.\n", remote_ip_str);

    return afd;
}

int direct_connect (struct gateway *gw, struct http_request *http, struct client_state *csp) {
    if (gw->forward_host)
        return (connect_to (gw->forward_host, gw->forward_port, csp));
    else
        return (connect_to (http->host, http->port, csp));
}

int connect_to (char *host, char *port, struct client_state * csp) {
    int fd = -1;
    int flags = -1;
    fd_set wfds;
    struct timeval tv[1];
    struct addrinfo hints, *res, *res0;
    int error;

    memset (&hints, 0, sizeof (hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = 0;

    error = getaddrinfo (host, port, &hints, &res0);
    if (error) {
        syslog (LOG_NOTICE, "Name lookup failed: %s port:%s", gai_strerror(error), port);
	printf("Name lookup failed: %s host:%s port:%s\n", gai_strerror(error), host, port);
        return (-1);
    }

    syslog (LOG_INFO, "Connecting to: %s at %s.\n", host, port);
    printf("Connecting to: %s at %s.\n", host, port);

    for (res = res0; res; res = res->ai_next) {
        fd = socket (res->ai_family, res->ai_socktype, res->ai_protocol);
        if (fd < 0)
            continue;

#ifdef TCP_NODELAY
        {                        /* turn off TCP coalescence */
            int mi = 1;
            setsockopt (fd, IPPROTO_TCP, TCP_NODELAY,
                        (char *) &mi, sizeof (int));
        }
#endif

        flags = fcntl (fd, F_GETFL, 0);
        if (flags != -1) {
            flags |= O_NDELAY;
            fcntl (fd, F_SETFL, flags);
        }

        error = connect (fd, (struct sockaddr *) res->ai_addr, res->ai_addrlen);
        if (error < 0) {
            if (errno != EINPROGRESS) {
                close (fd);
                fd = -1;
                continue;
            }
        }
        break;                        /* connected to a server */
    }

    if (res0)
        freeaddrinfo (res0);

    if (fd < 0)
        return -1;

    if (flags != -1) {
        flags &= ~O_NDELAY;
        fcntl (fd, F_SETFL, flags);
    }

    /* wait for connection to complete */
    FD_ZERO (&wfds);
    FD_SET (fd, &wfds);

    tv->tv_sec = 10;
    tv->tv_usec = 0;

    if (select (fd + 1, NULL, &wfds, NULL, tv) <= 0) {
        (void) close (fd);
        return (-1);
    }
    return (fd);
}

